
{***************************************************************************}
{                                                                           }
{     Unit: 'SpeechRecogWaveFile.pas'                                       }
{     ( containing class defining WaveFileStructure and methods )           }
{                                                                           }
{     - belonging to:                                                       }
{       Form:    'SpeechRecogForm.dfm'                                      }
{       Unit:    'SpeechRecogFormFile.pas'                                  }
{       Project: 'SpeechRecognition.dpr'                                    }
{       DLL:     'SpeechRecognition.dll'                                    }
{       Program: 'ResNeuronum.exe'                                          }
{       ( simulating Neural Networks )                                      }
{                                                                           }
{     - written in Borland Delphi 2.0                                       }
{                                                                           }
{     Copyright (c) 1996-97 TU Ilmenau/ Germany                             }
{                                                                           }
{     Written by:                                                           }
{       Christian Heller                                                    }
{       Grenzgraben 04                                                      }
{       D-98714 Sttzerbach                                                 }
{       Thringen                                                           }
{       Deutschland/ Germany/ Allemagne                                     }
{       Tel.:   +49/ (0)36784/ 50471 (Euro-ISDN)                            }
{       or Tel: +49/ (0)36784/ 50690                                        }
{       Fax G3: +49/ (0)36784/ 50992 (Euro-ISDN)                            }
{       e-mail: Christian.Heller@rz.tu-ilmenau.de                           }
{       www:    http://www.rz.tu-ilmenau.de/~ai008/christian/enmain.htm     }
{       www:    http://www.rz.tu-ilmenau.de/~ai008/resneuronum/enmain.htm   }
{                                                                           }
{     ( all variables & methods in alphabetical order                       }
{     or as appearing in the unit header, respectively )                    }
{                                                                           }
{***************************************************************************}

unit SpeechRecogWaveFile;

interface

uses
  Classes,
  Dialogs,
  { Unit for storing values in dynamic arrays }
  SpeechRecogDynArrayFile,
  SysUtils,
  Windows;

type
  { Definition of WaveFileFormat - in order of appearance in file header }
  { ! Do not change types of variables such as 'Integer' instead of 'Word' ! }
  TWaveFormat = record
    { Main chunk of file }
    RiffChunkLabel: array[0..3] of AnsiChar;
    RiffChunkLength: Cardinal;
    { File type }
    FileTypeLabel: array[0..3] of AnsiChar;
    { Format chunk = 1st section }
    FmtChunkLabel: array[0..3] of AnsiChar;
    FmtChunkLength: Cardinal;
    { Format type }
    wFormatTag: Word;
    { Number of channels }
    nChannels: Word;
    { Samples per second = sample rate }
    nSamplesPerSec: Cardinal;
    { Estimated value useful for dimensioning Buffer for output of sample }
    nAvgBytesPerSec: Cardinal;
    { Block alignment of sample data in file }
    nBlockAlign: Word;
    { Bits per sample in PCM-format-type }
    nBitsPerSample: Word;
    { Data chunk = 2nd section }
    DataChunkLabel: array[0..3] of AnsiChar;
    DataChunkLength: Cardinal;
  end;

  { Windows for SignalProcessing }
  TWindowType = set of (wtRectangle, wtHamming, wtHann, wtBartlett, wtBlackman);

  { Parameters for word detection }
  TWordDetParams = record
    PauseCount: Integer;
    SpeechLevel: Integer;
    WordBeginFrame: Integer;
    WordCount: Integer;
    WordEndFrame: Integer;
  end;

  { Class defining an object 'WaveFile' able to store its data in streams }
  TWaveFile = class(TObject)
  private
    FAbsMax: Double;
    FAverage: Double;
    FChosenFrame: Integer;
    FDataBark: TDoubleArray;
    FDataBarkFrame: TDoubleArray;
    FDataCent: TDoubleArray;
    FDataCutFrame: TDoubleArray;
    FDataLoud: TDoubleArray;
    FDataLoudFrame: TDoubleArray;
    FDataNorm: TDoubleArray;
    FDataPow: TDoubleArray;
    FDataPowFrame: TDoubleArray;
    FDataWave: TDoubleArray;
    FDataWindFrame: TDoubleArray;
    FFactor: Double;
    FFrameNo: Integer;
    FFrameWidth: Integer;
    FSampleNo: Integer;
    FSum: Double;
    FWindowType: TWindowType;
    FWaveFileName: TFileName;
    FWaveFormat: TWaveFormat;
    FWordDetParams: TWordDetParams;
  protected
    procedure ApplyMelScale(const APowFrameArray: TDoubleArray;
      var ABarkFrameArray: TDoubleArray; var ALoudFrameArray: TDoubleArray);
      virtual;
    procedure ApplyFFT(const AnInputArray: TDoubleArray;
      var AnOutputArray: TDoubleArray; const Size: Integer); virtual;
    procedure ApplyWindow(const AnInputArray: TDoubleArray;
      var AnOutputArray: TDoubleArray); virtual;
    procedure CalcLoudness;
    procedure CalcPower; virtual;
    procedure DetectWord; virtual;
    procedure LoadFromStream(FileStream: TFileStream); virtual;
    procedure Normalize; virtual;
    procedure SaveToStream(FileStream: TFileStream); virtual;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure LoadFromFile(const FileName: TFileName); virtual;
    procedure SaveToFile(const FileName: TFileName); virtual;
    procedure SignalProcessing; virtual;
    property AbsMax: Double read FAbsMax write FAbsMax;
    property Average: Double read FAverage write FAverage;
    property ChosenFrame: Integer read FChosenFrame write FChosenFrame;
    property DataBark: TDoubleArray read FDataBark write FDataBark;
    property DataBarkFrame: TDoubleArray read FDataBarkFrame write FDataBarkFrame;
    property DataCent: TDoubleArray read FDataCent write FDataCent;
    property DataCutFrame: TDoubleArray read FDataCutFrame write FDataCutFrame;
    property DataLoud: TDoubleArray read FDataLoud write FDataLoud;
    property DataLoudFrame: TDoubleArray read FDataLoudFrame write FDataLoudFrame;
    property DataNorm: TDoubleArray read FDataNorm write FDataNorm;
    property DataPow: TDoubleArray read FDataPow write FDataPow;
    property DataPowFrame: TDoubleArray read FDataPowFrame write FDataPowFrame;
    property DataWave: TDoubleArray read FDataWave write FDataWave;
    property DataWindFrame: TDoubleArray read FDataWindFrame write FDataWindFrame;
    property Factor: Double read FFactor write FFactor;
    property FrameNo: Integer read FFrameNo;
    property FrameWidth: Integer read FFrameWidth;
    property SampleNo: Integer read FSampleNo;
    property Sum: Double read FSum write FSum;
    property WindowType: TWindowType read FWindowType write FWindowType;
    property WaveFileName: TFileName read FWaveFileName write FWaveFileName;
    property WaveFormat: TWaveFormat read FWaveFormat;
    property WordDetParams: TWordDetParams read FWordDetParams
      write FWordDetParams;
  end;

implementation

{ FFT --> PowerSpectrum }
procedure TWaveFile.ApplyFFT(const AnInputArray: TDoubleArray;
  var AnOutputArray: TDoubleArray; const Size: Integer);
var
  a: Double;
  FFT_Values_Real, FFT_Values_Im: TDoubleArray;
  mr, i, j, l, iStep, el, m, nn: Integer;
  tr, ti, wr, wi: Double;
begin
  FFT_Values_Real := TDoubleArray.Create(Size + 1);
  FFT_Values_Im := TDoubleArray.Create(Size + 1);
  if ((FFT_Values_Real <> nil) and (FFT_Values_Im <> nil)) then begin
    for i := 1 to Size do begin
      FFT_Values_Real[i] := AnInputArray[i - 1];
      FFT_Values_Im[i] := 0.0;
    end;
    mr := 0;
    nn := (Size - 1);
    for m := 1 to nn do begin
      l := (Size div 2);
      while ((mr + l) > nn) do l := (l div 2);
      mr := ((mr mod l) + l);
      if (mr > m) then begin
        tr := FFT_Values_Real[m + 1];
        FFT_Values_Real[m + 1] := FFT_Values_Real[mr + 1];
        FFT_Values_Real[mr + 1] := tr;
        ti := FFT_Values_Im[m + 1];
        FFT_Values_Im[m + 1]  := FFT_Values_Im[mr + 1];
        FFT_Values_Im[mr + 1] := ti;
      end;
    end;
    l := 1;
    while not (l >= Size) do begin
      iStep := (2 * l);
      el := l;
      for m := 1 to l do begin
        a  := (Pi * (1 - m) / el);
        wr := cos(a);
        wi := sin(a);
        i  := m;
        while (i <= Size) do begin
          j  := (i + l);
          tr := ((wr * FFT_Values_Real[j]) - (wi * FFT_Values_Im[j]));
          ti := ((wr * FFT_Values_Im[j]) + (wi * FFT_Values_Real[j]));
          FFT_Values_Real[j] := (FFT_Values_Real[i] - tr);
          FFT_Values_Im[j] := (FFT_Values_Im[i] - ti);
          FFT_Values_Real[i] := (FFT_Values_Real[i] + tr);
          FFT_Values_Im[i] := (FFT_Values_Im[i] + ti);
          i := (i + iStep);
        end;
      end;
      l := iStep;
    end;
    for i := 0 to (Size - 1) do begin
      AnOutputArray[i] :=
      (Sqr(FFT_Values_Real[i + 1] / Size) + Sqr(FFT_Values_Im[i + 1] / Size));
    end;
    FFT_Values_Real.Free;
    FFT_Values_Im.Free;
  end
  { ... if not enough memory left }
  else begin
    if (FFT_Values_Real <> nil) then FFT_Values_Real.Free;
    if (FFT_Values_Im <> nil) then FFT_Values_Im.Free;
  end;
end;

procedure TWaveFile.ApplyMelScale(const APowFrameArray: TDoubleArray;
  var ABarkFrameArray: TDoubleArray; var ALoudFrameArray: TDoubleArray);

  { Transformate frequency from Hertz into Bark }
  function FrequBark(const FrequHz: Double): Double;
  begin
    try
      Result := (6 * ln((FrequHz / 600) + Sqrt(Sqr(FrequHz / 600) + 1)));
    except
      on Exception do Result := 0;
    end;
  end;

  { Involve Base to the Exponent }
  function Involve(const Base: Double; const Exponent: Double): Double;
  begin
    try
      Result := Exp(Ln(Base) * Exponent);
    except
      on Exception do Result := 0;
    end;
  end;

var
  i, k: Integer;
  f0: Double;
begin
  f0 := (11025 / FFrameWidth); // Delta between 256 frequencies in spectrum
  { Sum up frequencies in ranges specified by BarkScale }
  for k := 0 to (18 - 1) do begin
    ABarkFrameArray[k] := 0;
    ALoudFrameArray[k] := 0;
  end;
  for i := 0 to ((FFrameWidth div 2) - 1) do begin
    for k := 0 to (18 - 1) do begin
      if (((k + 1) - 1.3) <= FrequBark(i * f0)) and
      (FrequBark(i * f0) <= ((k + 1) - 0.5))
      then ABarkFrameArray[k] := (ABarkFrameArray[k] + (Involve(10,
      (2.5 * ((FrequBark(i * f0) - (k + 1)) + 0.5))) * APowFrameArray[i]));
      if (((k + 1) - 0.5) < FrequBark(i * f0)) and
      (FrequBark(i * f0) < ((k + 1) + 0.5))
      then ABarkFrameArray[k] := (ABarkFrameArray[k] + (1 * APowFrameArray[i]));
      if (((k + 1) + 0.5) <= FrequBark(i * f0)) and
      (FrequBark(i * f0) <= ((k + 1) + 2.5))
      then ABarkFrameArray[k] := (ABarkFrameArray[k] + (Involve(10,
      (-1 * ((FrequBark(i * f0) - (k + 1)) - 0.5))) * APowFrameArray[i]));
    end;
  end;
  { Calculate Loudness from 18 Bark frequencies }
  for k := 0 to (18 - 1) do begin
    ALoudFrameArray[k] := Involve(ABarkFrameArray[k], 0.5);
  end;
end;

procedure TWaveFile.ApplyWindow(const AnInputArray: TDoubleArray;
  var AnOutputArray: TDoubleArray);
var
  c: Double;
  j: Integer;
begin
  for j := 0 to (AnInputArray.Count - 1) do begin
    { Rectangle }
    if (FWindowType = [wtRectangle]) then begin
      { Leave values unchanged }
    end;
    { Hamming = Raised Cosinus Window }
    if (FWindowType = [wtHamming]) then begin
      c := 0.54;
      AnOutputArray[j] := (AnInputArray[j] * (c - ((1 - c) *
      cos(2 * Pi * j / AnInputArray.Count))));
    end;
    { Hann = Hanning Window }
    if (FWindowType = [wtHann]) then begin
      c := 0.5;
      AnOutputArray[j] := (AnInputArray[j] * (c - ((1 - c) *
      cos(2 * Pi * j / AnInputArray.Count))));
    end;
    { Bartlett Triangle Window }
    if (FWindowType = [wtBartlett]) then begin
      if (j < (AnInputArray.Count / 2))
      then AnOutputArray[j] := (AnInputArray[j] *
      (2 * j / AnInputArray.Count))
      else AnOutputArray[j] := (AnInputArray[j] *
      (2 - (2 * j / AnInputArray.Count)));
    end;
    { Blackman Window }
    if (FWindowType = [wtBlackman]) then begin
      c := 0.42;
      AnOutputArray[j] := (AnInputArray[j] *
      (c - (0.5 * cos(2 * Pi * j / AnInputArray.Count)) +
      ((0.5 - c) * cos(4 * Pi * j / AnInputArray.Count))));
    end;
  end;
end;

procedure TWaveFile.CalcLoudness;
var
  i, j, k: Integer;
  DataCutLocal: TDoubleArray;
  DataWindLocal: TDoubleArray;
  DataPowLocal: TDoubleArray;
  DataBarkLocal: TDoubleArray;
  DataLoudLocal: TDoubleArray;
begin
  DataCutLocal := TDoubleArray.Create(0);
  DataWindLocal := TDoubleArray.Create(0);
  DataPowLocal := TDoubleArray.Create(0);
  DataBarkLocal := TDoubleArray.Create(0);
  DataLoudLocal := TDoubleArray.Create(0);
  { Calculate only for frames within previously detected word }
  with FWordDetParams do for i := (WordBeginFrame - 1) to (WordEndFrame - 1) do begin
    { DataCutLocal }
    DataCutLocal.Free;
    DataCutLocal := TDoubleArray.Create(FFrameWidth);
    for j := 0 to (FrameWidth - 1) do begin
      DataCutLocal[j] := FDataNorm[(i * FFrameWidth) + j];
    end;
    { DataWindLocal }
    DataWindLocal.Free; DataWindLocal := nil;
    DataWindLocal := TDoubleArray.Create(FFrameWidth);
    ApplyWindow(DataCutLocal, DataWindLocal);
    { DataPowLocal }
    DataPowLocal.Free; DataPowLocal := nil;
    DataPowLocal := TDoubleArray.Create(FFrameWidth);
    ApplyFFT(DataWindLocal, DataPowLocal, DataWindLocal.Count);
    { DataBarkLocal and DataLoudLocal }
    DataBarkLocal.Free; DataBarkLocal := nil;
    DataBarkLocal := TDoubleArray.Create(18);
    DataLoudLocal.Free; DataLoudLocal := nil;
    DataLoudLocal := TDoubleArray.Create(18);
    ApplyMelScale(DataPowLocal, DataBarkLocal, DataLoudLocal);
    { Sum up each of the 18 logarithmic bands over the whole signal }
    for k := 0 to (18 - 1) do begin
      with FWordDetParams do begin
        FDataBark[k] := (FDataBark[k] + (DataBarkLocal[k] / (WordEndFrame - WordBeginFrame)));
        FDataLoud[k] := (FDataLoud[k] + (DataLoudLocal[k] / (WordEndFrame - WordBeginFrame)));
      end;
    end;
    { Collect data for diagram }
    if (i = (FChosenFrame - 1)) then begin
      for k := 0 to (DataBarkLocal.Count - 1) do begin
        FDataBarkFrame[k] := DataBarkLocal[k];
        FDataLoudFrame[k] := DataBarkLocal[k];
      end;
    end;
  end;
  DataCutLocal.Free;
  DataWindLocal.Free;
  DataPowLocal.Free;
  DataBarkLocal.Free;
  DataLoudLocal.Free;
end;

procedure TWaveFile.CalcPower;
var
  i, j: Integer;
  DataCutLocal: TDoubleArray;
  DataWindLocal: TDoubleArray;
  DataPowLocal: TDoubleArray;
begin
  DataCutLocal := TDoubleArray.Create(0);
  DataWindLocal := TDoubleArray.Create(0);
  DataPowLocal := TDoubleArray.Create(0);
  for i := 0 to (FFrameNo - 1) do begin
    { DataCutFrame }
    DataCutLocal.Free;
    DataCutLocal := TDoubleArray.Create(FFrameWidth);
    for j := 0 to (FrameWidth - 1) do begin
      DataCutLocal[j] := FDataNorm[(i * FFrameWidth) + j];
    end;
    { DataWindFrame }
    DataWindLocal.Free; DataWindLocal := nil;
    DataWindLocal := TDoubleArray.Create(FFrameWidth);
    ApplyWindow(DataCutLocal, DataWindLocal);
    { DataPowFrame }
    DataPowLocal.Free; DataPowLocal := nil;
    DataPowLocal := TDoubleArray.Create(FFrameWidth);
    ApplyFFT(DataWindLocal, DataPowLocal, DataWindLocal.Count);
    { Sum up all 128 power values (half info is redundant) for each time }
    FDataPow[i] := 0;
    for j := 0 to ((FFrameWidth div 2) - 1) do
      FDataPow[i] := (FDataPow[i] + (DataPowLocal[j] / (FFrameWidth div 2)));
    { Collect data for diagram }
    if (i = (FChosenFrame - 1)) then for j := 0 to (FrameWidth - 1) do begin
      FDataCutFrame[j] := DataCutLocal[j];
      FDataWindFrame[j] := DataWindLocal[j];
      FDataPowFrame[j] := DataPowLocal[j];
    end;
  end;
  DataCutLocal.Free;
  DataWindLocal.Free;
  DataPowLocal.Free;
end;

constructor TWaveFile.Create;
begin
  inherited Create;
  with FWaveFormat do begin
    { Main chunk of file }
    StrCopy(RiffChunkLabel, 'RIFF');
    RiffChunkLength := 36;
    { File type }
    StrCopy(FileTypeLabel, 'WAVE');
    { Format chunk = 1st section }
    StrCopy(FmtChunkLabel, 'fmt ');
    FmtChunkLength := 16;
    { Format type PCM }
    wFormatTag := 1;
    { Number of channels }
    nChannels := 1;
    { Samples per second = sample rate }
    nSamplesPerSec := 11025;
    { Estimated value useful for dimensioning Buffer for output of sample }
    nAvgBytesPerSec := 22050;
    { Block alignment of sample data in file }
    nBlockAlign := 2;
    { Bits per sample in PCM-format-type }
    nBitsPerSample := 16;
    { Data chunk = 2nd section }
    StrCopy(DataChunkLabel, 'data');
    DataChunkLength := 0;
  end;
  { Create dynamic arrays }
  FDataBark := TDoubleArray.Create(0);
  FDataBarkFrame := TDoubleArray.Create(0);
  FDataCent := TDoubleArray.Create(0);
  FDataCutFrame := TDoubleArray.Create(0);
  FDataLoud := TDoubleArray.Create(0);
  FDataLoudFrame := TDoubleArray.Create(0);
  FDataNorm := TDoubleArray.Create(0);
  FDataPow := TDoubleArray.Create(0);
  FDataPowFrame := TDoubleArray.Create(0);
  FDataWave := TDoubleArray.Create(0);
  FDataWindFrame := TDoubleArray.Create(0);
  { Initiate various variables }
  FAbsMax := 0;
  FAverage := 0;
  FFactor := 32767;
  FFrameNo := 0;
  FFrameWidth := 256;
  FSampleNo := 0;
  FSum := 0;
  FWaveFileName := 'Untitled';
end;

destructor TWaveFile.Destroy;
begin
  FDataBark.Free; FDataBark := nil;
  FDataBarkFrame.Free; FDataBarkFrame := nil;
  FDataCent.Free; FDataCent := nil;
  FDataCutFrame.Free; FDataCutFrame := nil;
  FDataLoud.Free; FDataLoud := nil;
  FDataLoudFrame.Free; FDataLoudFrame := nil;
  FDataNorm.Free; FDataNorm := nil;
  FDataPow.Free; FDataPow := nil;
  FDataPowFrame.Free; FDataPowFrame := nil;
  FDataWave.Free; FDataWave := nil;
  FDataWindFrame.Free; FDataWindFrame := nil;
  inherited Destroy;
end;

procedure TWaveFile.DetectWord;
var
  Counter: Integer;
  i: Integer;
begin
  Counter := 0;
  with FWordDetParams do begin
    WordBeginFrame := -1;
    WordEndFrame := -1;
    for i := 0 to (FrameNo - 1) do begin
      { Increase Counter if speech level crossed and no word finished before }
      if (FDataPow[i] > SpeechLevel) and (Counter < PauseCount) then Inc(Counter);
      { Decrease Counter if SpeechLevel is undercut and Counter not zero yet }
      if (FDataPow[i] <= SpeechLevel) and (Counter > 0) then Dec(Counter);
      { Detect begin of word }
      if ((Counter > WordCount) and (WordBeginFrame < 0)) then begin
        WordBeginFrame := ((i + 1) - WordCount);
        Counter := PauseCount;
      end;
      { Detect end of word }
      if ((Counter = 0) and (WordEndFrame < 0) and (WordBeginFrame > -1)) then begin
        WordEndFrame := ((i + 1) - PauseCount);
        Exit;
      end;
    end;
    { Prevent errors if no word has been detected }
    if (WordBeginFrame < 0) then WordBeginFrame := 1;
    if ((WordEndFrame < 0) or (WordEndFrame > FrameNo)) then WordEndFrame := FrameNo;
  end;
end;

procedure TWaveFile.LoadFromFile(const FileName: TFileName);
var
  FileStream: TFileStream;
begin
  FileStream := TFileStream.Create(FileName, fmOpenRead);
  try
    { Call internal stream procedure to read from file }
    LoadFromStream(FileStream);
    { Set name of active wave file }
    FWaveFileName := FileName;
    { SignalProcessing to get spectra etc. }
    SignalProcessing;
  finally
    FileStream.Free;
  end;
end;

procedure TWaveFile.LoadFromStream(FileStream: TFileStream);
var
  i: Integer;
  { Sample has to be SmallInt with 2 Bytes for compatibility with data! }
  Sample: SmallInt;
begin
  with FWaveFormat do begin
    with FileStream do begin
      { Main chunk of file }
      ReadBuffer(RiffChunkLabel, SizeOf(RiffChunkLabel));
      ReadBuffer(RiffChunkLength, SizeOf(RiffChunkLength));
      { File type }
      ReadBuffer(FileTypeLabel, SizeOf(FileTypeLabel));
      { Format chunk = 1st section }
      ReadBuffer(FmtChunkLabel, SizeOf(FmtChunkLabel));
      ReadBuffer(FmtChunkLength, SizeOf(FmtChunkLength));
      { Format type }
      ReadBuffer(wFormatTag, SizeOf(wFormatTag));
      { Number of channels }
      ReadBuffer(nChannels, SizeOf(nChannels));
      { Samples per second = sample rate }
      ReadBuffer(nSamplesPerSec, SizeOf(nSamplesPerSec));
      { Estimated value useful for dimensioning Buffer for output of sample }
      ReadBuffer(nAvgBytesPerSec, SizeOf(nAvgBytesPerSec));
      { Block alignment of sample data in file }
      ReadBuffer(nBlockAlign, SizeOf(nBlockAlign));
      { Bits per sample in PCM-format-type }
      ReadBuffer(nBitsPerSample, SizeOf(nBitsPerSample));
      { Data chunk = 2nd section }
      ReadBuffer(DataChunkLabel, SizeOf(DataChunkLabel));
      ReadBuffer(DataChunkLength, SizeOf(DataChunkLength));
    end;
    { Check file format }
    if (RiffChunkLabel <> 'RIFF') or
    (FileTypeLabel <> 'WAVE') or
    (FmtChunkLabel <> 'fmt ') or
    (wFormatTag <> 1) or
    (nChannels <> 1) or
    (nSamplesPerSec <> 11025) or
    (DataChunkLabel <> 'data')
    then begin
      MessageBeep(0);
      raise Exception.Create('Unbekanntes Dateiformat!');
    end;
    if (DataChunkLength > 0) then begin
      { Calculate number of samples }
      FSampleNo := (DataChunkLength div SizeOf(Sample));
      { Calculate FrameNo that the given wave file consists of }
      FFrameNo := (FSampleNo div FFrameWidth);
      { Free existing dynamic WaveData array and create new one }
      FDataWave.Free; FDataWave := nil;
      FDataWave := TDoubleArray.Create(FSampleNo);
      if (FDataWave <> nil) then begin
        { Read values from file and store them in dynamic array }
        for i := 0 to (FSampleNo - 1) do begin
          FileStream.ReadBuffer(Sample, SizeOf(Sample));
          FDataWave[i] := Sample;
        end;
      end
      else begin
        MessageBeep(0);
        raise Exception.Create('Nicht genug Speicher vorhanden!');
      end;
    end
    else begin
      MessageBeep(0);
      raise Exception.Create('Die Wave-Datei enthlt keine Daten!');
    end;
  end;
end;

procedure TWaveFile.Normalize;
var
  i: Integer;
begin
  { Get sum and maximum of signal }
  FSum := 0;
  FAbsMax := 0;
  for i := 0 to (FSampleNo - 1) do begin
    FSum := (FSum + FDataWave[i]);
    if (Abs(FDataWave[i]) > FAbsMax) then FAbsMax := Abs(FDataWave[i]);
  end;
  try
    { Calculate average of data }
    FAverage := FSum / FSampleNo;
    try
      { Calculate factor that relates values to absolute maximum of volume }
      FFactor := (32767 / (FAbsMax - FAverage)); // Maximum of WaveData = 32767
    except
      on Exception do FFactor := 32767;
    end;
  except
    on Exception do FAverage := 0;
  end;
  for i := 0 to (FSampleNo - 1) do begin
    { Subtract equal portion of signal, e.g. voltage caused by microphone }
    FDataCent[i] := FDataWave[i] - FAverage; // Centralization
    { Relate to MaxVolume by factor to compensate different record levels }
    FDataNorm[i] := (FDataCent[i] * FFactor); // Normalization
  end;
end;

procedure TWaveFile.SaveToFile(const FileName: TFileName);
var
  FileStream: TFileStream;
begin
  { Create FileStream }
  FileStream := TFileStream.Create(FileName, fmCreate);
  try
    { Call internal Stream-procedure to write to file }
    SaveToStream(FileStream);
  finally
    FileStream.Free;
  end;
end;

procedure TWaveFile.SaveToStream(FileStream: TFileStream);
begin
  with WaveFormat do begin
    with FileStream do begin
      { Main chunk of file }
      WriteBuffer(RiffChunkLabel, SizeOf(RiffChunkLabel));
      WriteBuffer(RiffChunkLength, SizeOf(RiffChunkLength));
      { File type }
      WriteBuffer(FileTypeLabel, SizeOf(FileTypeLabel));
      { Format chunk = 1st section }
      WriteBuffer(FmtChunkLabel, SizeOf(FmtChunkLabel));
      WriteBuffer(FmtChunkLength, SizeOf(FmtChunkLength));
      { Format type }
      WriteBuffer(wFormatTag, SizeOf(wFormatTag));
      { Number of channels }
      WriteBuffer(nChannels, SizeOf(nChannels));
      { Samples per second = sample rate }
      WriteBuffer(nSamplesPerSec, SizeOf(nSamplesPerSec));
      { Estimated value useful for dimensioning Buffer for output of sample }
      WriteBuffer(nAvgBytesPerSec, SizeOf(nAvgBytesPerSec));
      { Block alignment of sample data in file }
      WriteBuffer(nBlockAlign, SizeOf(nBlockAlign));
      { Bits per sample in PCM-format-type }
      WriteBuffer(nBitsPerSample, SizeOf(nBitsPerSample));
      { Data chunk = 2nd section }
      WriteBuffer(DataChunkLabel, SizeOf(DataChunkLabel));
      WriteBuffer(DataChunkLength, SizeOf(DataChunkLength));
    end;
  end;
end;

procedure TWaveFile.SignalProcessing;
var
  i, k: Integer;
begin
  if (FSampleNo > 0) then begin
    { Section 'Centralization' and 'Normalization' }
    FDataCent.Free; FDataCent := nil;
    FDataCent := TDoubleArray.Create(FSampleNo);
    FDataNorm.Free; FDataNorm := nil;
    FDataNorm := TDoubleArray.Create(FSampleNo);
    Normalize;
    { Section 'Window' and 'FFT' }
    FDataPow.Free; FDataPow := nil;
    FDataPow := TDoubleArray.Create(FFrameNo);
    FDataCutFrame.Free; FDataCutFrame := nil;
    FDataCutFrame := TDoubleArray.Create(FFrameWidth);
    FDataWindFrame.Free; FDataWindFrame := nil;
    FDataWindFrame := TDoubleArray.Create(FFrameWidth);
    FDataPowFrame.Free; FDataPowFrame := nil;
    FDataPowFrame := TDoubleArray.Create(FFrameWidth);
    for i := 0 to (FrameWidth - 1) do begin
      FDataCutFrame[i] := 0;
      FDataWindFrame[i] := 0;
      FDataPowFrame[i] := 0;
    end;
    CalcPower;
    { Section 'WordDetection' }
    DetectWord;
    { Section 'BarkScaling' and 'Loudness' }
    FDataBark.Free; FDataBark := nil;
    FDataBark := TDoubleArray.Create(18);
    FDataLoud.Free; FDataLoud := nil;
    FDataLoud := TDoubleArray.Create(18);
    FDataBarkFrame.Free; FDataBarkFrame := nil;
    FDataBarkFrame := TDoubleArray.Create(18);
    FDataLoudFrame.Free; FDataLoudFrame := nil;
    FDataLoudFrame := TDoubleArray.Create(18);
    for k := 0 to (18 - 1) do begin
      FDataBark[k] := 0;
      FDataLoud[k] := 0;
    end;
    CalcLoudness;
  end;
end;

end.
